/*****************************************************************************************
 * Copyright (C), 2004
 * SAS Institute Inc., Cary, N.C. 27513, U.S.A.  All rights reserved.
 * 
 *  NAME:    ORVA_scaling_internal                                            
 *  TITLE:   Creates a scaling dataset to go along with a scaling component.
 *           Generally called by orva_scaling() when dataset has been reset.
 *
 *  SUPPORT: stchri
 *
 *  PARAMETERS:
 *     varcomp  - libref for the OpRiskVar Components directory.
 *     varorg   - libref for the OpRiskVar Organizational Structure directory.
 *     varmeta  - libref for the OpRiskVar Meta directory.
 *     scaleUid - UID for the Scaling Component to create dataset for.
 *     scaleDS  - output dataset.
 *      
 *****************************************************************************************/
%macro orva_scaling_internal(
                              varcomp  =,
                              varorg   =,
                              varmeta  =,
                              scaleUid =,
                              scaleDS  =
                              );

  %put %sysfunc(sasmsg(&orvads, status_note, noquote)) %sysfunc(sasmsg(&orvads, sint_call_note, noquote, &scaleDS ));
                             
  /*****- get various macro variables for the scaling object -*****/
  %let scaleFound = false;
  data _NULL_; set &varcomp..internal_scale_adjustment;
      where INTERNAL_SCALE_ADJUSTMENT_UID=&scaleUid;
      call symput("scaleFound", 'true');
      call symput("freqScaleFirm", FREQUENCY_SCALE_BY_FIRM);
      call symput("freqScaleBl", FREQUENCY_SCALE_BY_BL);
      call symput("sevScaleFirm", SEVERITY_SCALE_BY_FIRM);
      call symput("sevScaleBl", SEVERITY_SCALE_BY_BL);
      call symput("freqCoefFirm", FREQUENCY_COEFFICIENT_FIRM);
      call symput("freqCoefBl", FREQUENCY_COEFFICIENT_BL);
      call symput("sevCoefFirm", SEVERITY_COEFFICIENT_FIRM);
      call symput("sevCoefBl", SEVERITY_COEFFIECIENT_BL);
      if FREQUENCY_SCALE_BY_FIRM eq 0 then FREQUENCY_SFV_FIRM = -1;
      call symput("freqSfvFirm", FREQUENCY_SFV_FIRM);
      if FREQUENCY_SCALE_BY_BL eq 0 then FREQUENCY_SFV_BL = -1;
      call symput("freqSfvBl", FREQUENCY_SFV_BL);
      if SEVERITY_SCALE_BY_FIRM eq 0 then SEVERITY_SFV_FIRM = -1;
      call symput("sevSfvFirm", SEVERITY_SFV_FIRM);
      if SEVERITY_SCALE_BY_BL eq 0 then SEVERITY_SFV_BL = -1;
      call symput("sevSfvBl", SEVERITY_SFV_BL);
  run;
  
  %if &scaleFound eq false %then %do;
     %if &scaleUid GE 1000 %then %do; 
        %put &orvar_warning %sysfunc(sasmsg(&orvads, sint_adj_not_found_error, noquote, &scaleUid));
     %end;
     %return;
  %end;
  
  data _NULL_; set &varorg..global_range;
    call symput("startYear", year(START_DATE));
    call symput("endYear", year(END_DATE));
  run;
  
  /*****- handle Frequency   -*****/ 
  %if &freqScaleFirm ne 0 or &freqScaleBl ne 0 %then %do;
     %getScaleFactors(scaleUid=&scaleUid, varcomp=&varcomp, varorg=&varorg, varmeta=&varmeta,
                      sfvFirm=&freqSfvFirm, sfvBl=&freqSfvBl, out=_freqValues, type=freq,
                      startYear=&startYear, endYear=&endYear);
     %adjustUsingAgr(data=_freqValues,type=freq);   
     %adjustNotSelected(data=_freqValues,type=freq); 
     %calcSaf(data=_freqValues,type=freq,coefFirm=&freqCoefFirm,coefBl=&freqCoefBl);
     data _freqValues0; set _freqValues;
        BLUID_SC = internal_bl_uid;
        FSCALE = saf_final;
     keep BLUID_SC YEAR FSCALE;
     run;
     proc sort data = _freqValues0; by BLUID_SC YEAR; run;
  %end;
  %else %do;
     proc sql; create table _freqValues0 (
       BLUID_SC num,
       YEAR num,
       FSCALE num);
     quit;
  %end;

  /*****- handle Severity -*****/
  %if &sevScaleFirm ne 0 or &sevScaleBl ne 0 %then %do;
     %getScaleFactors(scaleUid=&scaleUid, varcomp=&varcomp, varorg=&varorg, varmeta=&varmeta,
                      sfvFirm=&sevSfvFirm, sfvBl=&sevSfvBl, out=_sevValues, type=sev,
                      startYear=&startYear, endYear=&endYear);
     %adjustUsingAgr(data=_sevValues,type=sev);
     %adjustNotSelected(data=_sevValues,type=sev);
     %calcSaf(data=_sevValues,type=sev,coefFirm=&sevCoefFirm,coefBl=&sevCoefBl);

     data _sevValues0; set _sevValues;
        BLUID_SC = internal_bl_uid;
        SSCALE_INIT = saf_final;
        SSCALE = saf_final;
        keep BLUID_SC YEAR SSCALE_INIT SSCALE;
     run;
     proc sort data = _sevValues0; by BLUID_SC YEAR; run;
  %end; 
  %else %do;
     proc sql; create table _sevValues0 (
        BLUID_SC num,
        YEAR num,
        SSCALE_INIT num,
        SSCALE num);
     quit;
  %end;

  /*****- merge severity and frequency together -*****/
  data &scaleDS; merge _sevValues0 _freqValues0; by BLUID_SC YEAR; run;

  /*****- cleanup temporary datasets -*****/
  proc datasets library=work memtype=data nolist nodetails;
      delete _freqValues: _sevValues:;
  quit;   
%mend orva_scaling_internal;

/***********************************************************************************
 * Apply Scaling formulas:
 *
 * FREQUENCY:
 *   saf_firm = (scaleValueFirm_ModelingYear / scaleValueFirm )** coefFirm * aaf_firm;
 *   saf_bl = (scaleValueBl_ModelingYear / scaleValueBl )** coefBl * aaf_bl;
 *
 * SEVERITY:
 *   saf_firm = (scaleValueFirm_ModelingYear / scaleValueFirm )** coefFirm;
 *   saf_bl = (scaleValueBl_ModelingYear / scaleValueBl )** coefBl;
 *
 * FINAL:
 *   saf = saf_firm * saf_bl
 **********************************************************************************/
%macro calcSaf(data=, type=, coefFirm=, coefBl=);  
     proc sort data=&data; by internal_bl_uid DESCENDING year; run;
     data &data; set &data end=last; 
        by internal_bl_uid DESCENDING year;
        retain scaleValueBl;
        retain scaleValueFirm;
        %if &type eq freq %then %do; 
          if (aaf_firm eq . or aaf_firm eq 0) then aaf_firm = 1;
          if (aaf_bl eq . or aaf_bl eq 0) then aaf_bl = 1;
        %end;
        if first.internal_bl_uid then do;
           scaleValueBl = &type._sfv_bl;
           scaleValueFirm = &type._sfv_firm;
        end; 
        saf_f = (scaleValueFirm / &type._sfv_firm )** &coefFirm;
        %if &type eq freq %then %do; 
           saf_f = saf_f * aaf_firm;
        %end;
        saf_bl = (scaleValueBl / &type._sfv_bl )** &coefBl;
        %if &type eq freq %then %do;
           saf_bl = saf_bl * aaf_bl;
        %end;
        if (saf_f ne 0 and saf_f ne .) then do;
           saf_final = saf_f;
           if saf_bl ne 0 and saf_bl ne . then saf_final = saf_final * saf_bl;
        end;
        else if saf_bl ne 0 and saf_bl ne . then saf_final = saf_bl;
        else saf_final = 1;
        drop scaleValueBl scaleValueFirm; 
     run;
%mend calcSaf;

/***********************************************************************************
 * After all other adjustments are made, reset sfv values when the user specifically
 * deselected scaling (by BL-Year or by Firm-Year as well as by Type)
 **********************************************************************************/
%macro adjustNotSelected(data=, type=);
   data &data; set &data;
      if selected_&type._bl ne 1 then do;
         &type._sfv_bl = .;
      end;
      if selected_&type._firm ne 1 then do;
         &type._sfv_firm = .;
      end;
      drop selected_&type._:;
   run;
%mend adjustNotSelected;

/***********************************************************************************
 * uses AGR to adjust missing values forward and backward
 *
 * FOWARD:
 *    saf = prev_saf + prev_saf * agr/100;
 *
 * BACKWARD:
 *    saf = next_saf / (1.0 + agr/100);
 *
 **********************************************************************************/ 
%macro adjustUsingAgr(data=, type=);
      /* adjust forward */
      proc sort data=&data; by internal_bl_uid year; run;
      data &data; set &data end=last; by internal_bl_uid year;
         retain prevFirm;
         retain prevBl;
         retain agrFirm;
         retain agrBl;
         if first.internal_bl_uid then do;
           prevFirm = &type._sfv_firm;
           prevBl = &type._sfv_bl;
           agrFirm = &type._agr_firm;
           if agrFirm eq . then agrFirm = 0;
           agrBl = &type._agr_bl;
           if agrBl eq . then agrBl = 0;
         end;
         if &type._agr_firm eq . then &type._agr_firm = agrFirm;
         if &type._agr_bl eq . then &type._agr_bl = agrBl;
         if first.internal_bl_uid eq 0 then do;
           if &type._sfv_firm eq . and prevFirm ne . then do;
              &type._sfv_firm = prevFirm + prevFirm * &type._agr_firm/100;
           end;
           if &type._sfv_bl eq . and prevBl ne .  then do;
              &type._sfv_bl = prevBl + prevBl * &type._agr_bl/100;
           end;
           prevFirm = &type._sfv_firm;
           prevBl = &type._sfv_bl;
         end;
         output;
         if last.internal_bl_uid then do;
           do while ( year < &endYear );
              aaf_bl = 1;
              aaf_firm = 1;
              year = year+1;
              if prevFirm ne . then do;
                 &type._sfv_firm = prevFirm + prevFirm * &type._agr_firm/100;
              end;
              if prevBl ne . then do;
                 &type._sfv_bl = prevBl + prevBl * &type._agr_bl/100;
              end;
              prevFirm = &type._sfv_firm;
              prevBl = &type._sfv_bl;
              output;
           end;
         end; 
         drop prevFirm prevBl;
       run;

      /* adjust backward */
      proc sort data=&data; by internal_bl_uid DESCENDING year; run;
      data &data; set &data end=last; by internal_bl_uid DESCENDING year;
         retain prevFirm;
         retain prevBl;
         if &type._agr_firm eq . then &type._agr_firm = 0;
         if &type._agr_bl eq . then &type._agr_bl = 0;
         if first.internal_bl_uid then do;
             prevFirm = &type._sfv_firm;
             prevBl = &type._sfv_bl;
         end;
         else do;
             if &type._sfv_firm eq . and prevFirm ne . then do;
               &type._sfv_firm = prevFirm / (1.0+&type._agr_firm/100);
             end;
             if &type._sfv_bl eq . and prevBl ne .  then do;
               &type._sfv_bl = prevBl / (1.0+&type._agr_bl/100);
             end;
             prevFirm = &type._sfv_firm;
             prevBl = &type._sfv_bl;
         end; 
         drop prevFirm prevBl;
      run;

%mend adjustUsingAgr;

/***********************************************************************************
 *
 *  1. get the default scale factor values from scale_factor_firm or scale_factor_bl.
 *  2. override with values from internal_<freq/sev>_adj_firm or internal_<freq/sev>_adj_bl.
 *  3. merge firm and businessline values into one dataset by year.
 *  4. At this point ignore the selected flag, will be taken care of later
 *     Just make sure to keep it around.
 **********************************************************************************/
%macro getScaleFactors(scaleUid=,
                       out=,
                       varcomp=,
                       varorg=,
                       varmeta=,
                       sfvFirm=,
                       sfvBl=,
                       type=,
                       startYear=,
                       endYear=);

   /****************************************/
   /********** handle FIRM values **********/
   /****************************************/
   %if &sfvFirm ^= -1 %then %do;  
      /*- get default scale factor values from SCALE_FACTOR_FIRM table -*/
      data _defaults; set &varorg..scale_factor_firm;
         where SCALE_FACTOR_VARIABLE_UID eq &sfvFirm and
               SCALE_FACTOR_TYPE eq 'F';
         &type._default = SCALE_FACTOR_VALUE;
         year = SCALE_FACTOR_YEAR;
         keep year &type._default;
      run;

      /*- get overidden FREQUENCY values (for &scaleUid) -*/
      %if &type eq freq %then %do;
         %let AGR=0;
         data _NULL_; set &varcomp..internal_frequency_agr_firm;
            where SCALE_FACTOR_VARIABLE_UID eq &sfvFirm and
                  INTERNAL_SCALE_ADJUSTMENT_UID = &scaleUid;
            call symput("AGR", INTERNAL_FREQ_AGR_VALUE);
         run;
         data _values; set &varcomp..internal_frequency_adj_firm;
            where SCALE_FACTOR_VARIABLE_UID eq &sfvFirm and
                 INTERNAL_SCALE_ADJUSTMENT_UID = &scaleUid;
            freq_override = SCALE_FACTOR_VALUE;
            year = INTERNAL_FREQ_ADJ_YEAR;
            aaf_firm = INTERNAL_FREQ_ADJ_AAF;
            keep year aaf_firm freq_override selected
        run;
      %end;

      /* get overidden SEVERITY values (for &scaleUid) */
      %else %do;
         %let AGR=0;
         data _NULL_; set &varcomp..internal_severity_agr_firm;
            where SCALE_FACTOR_VARIABLE_UID eq &sfvFirm and
                  INTERNAL_SCALE_ADJUSTMENT_UID = &scaleUid;
            call symput("AGR", INTERNAL_SEV_AGR_VALUE);
         run;
         data _values; set &varcomp..internal_severity_adj_firm;
            where SCALE_FACTOR_VARIABLE_UID eq &sfvFirm and
                 INTERNAL_SCALE_ADJUSTMENT_UID = &scaleUid;
            sev_override = SCALE_FACTOR_VALUE;
            year = INT_SEV_ADJ_YEAR;
            keep year sev_override selected;
         run;
      %end;

      /* merge overides and defaults together */
      proc sort data= _defaults; by year; run;
      proc sort data= _values; by year; run;
      data _valuesFirm; merge _values _defaults; by year;
          if &type._override ne -1 and &type._override ne . then 
             &type._sfv_firm = &type._override;
          else &type._sfv_firm = &type._default;
          selected_&type._firm = selected;
          keep year aaf_firm &type._sfv_firm selected_&type._firm;
      run;

      %fillin_levels_firm(data=_valuesFirm, varorg=&varorg,
              startYear=&startYear, endYear=&endYear,
              type=&type, agr=&AGR);

      proc sort data= _valuesFirm; by year internal_bl_uid; run;
      %if &sfvBl eq -1 %then %do;
         data &out; set _valuesFirm; 
         run; 
      %end;

  %end;

  /**************************************/
  /***** handle BusinessLine values *****/
  /**************************************/
  %if &sfvBl ^= -1 %then %do;
  
     /* get defaults from SCALE_FACTOR_BL table */
     data _defaults; set &varorg..scale_factor_bl;
         where SCALE_FACTOR_VARIABLE_UID eq &sfvBl;
         &type._default = SCALE_FACTOR_VALUE;
         year = SCALE_FACTOR_YEAR;
         keep year internal_bl_uid &type._default;
     run;
    
     /* make sure we only have Level-1 Businesslines */
     data _bl1; set &varorg..internal_businessline; 
         where internal_bl_level eq 1; 
         keep internal_bl_uid internal_bl_level;
     run;
     proc sort data=_defaults; by internal_bl_uid; run;
     proc sort data=_bl1;  by internal_bl_uid; run;
     data _defaults; merge _defaults _bl1; 
         by internal_bl_uid; 
         if internal_bl_level eq 1 then output;
     run;
     
     /* get overidden values FREQUENCY (for &scaleUid) */
     %if &type eq freq %then %do;
        data _agrs; set &varcomp..internal_frequency_agr_bl;
           where SCALE_FACTOR_VARIABLE_UID eq &sfvBl and
                 INTERNAL_SCALE_ADJUSTMENT_UID = &scaleUid;
           freq_agr_bl = INTERNAL_FREQ_AGR_VALUE;
           keep internal_bl_uid freq_agr_bl;
       run;
       data _values; set &varcomp..internal_frequency_adj_bl;
          where SCALE_FACTOR_VARIABLE_UID eq &sfvBl and
                INTERNAL_SCALE_ADJUSTMENT_UID = &scaleUid;
           freq_override = SCALE_FACTOR_VALUE;
           year = INTERNAL_FREQ_ADJ_YEAR;
           aaf_bl = INTERNAL_FREQ_ADJ_AAF;
           keep year internal_bl_uid aaf_bl freq_override selected;
       run;
     %end;
     /* get overidden values SEVERITY (for &scaleUid) */
     %else %do;
        data _agrs; set &varcomp..internal_severity_agr_bl;
          where SCALE_FACTOR_VARIABLE_UID eq &sfvBl and
                INTERNAL_SCALE_ADJUSTMENT_UID = &scaleUid;
            sev_agr_bl = INTERNAL_SEV_AGR_VALUE;
            keep internal_bl_uid sev_agr_bl;
       run;
       data _values; set &varcomp..internal_severity_adj_bl;
          where SCALE_FACTOR_VARIABLE_UID eq &sfvBl and
                INTERNAL_SCALE_ADJUSTMENT_UID = &scaleUid;
          sev_override = SCALE_FACTOR_VALUE;
          year = INT_SEV_ADJ_YEAR;
          keep year internal_bl_uid sev_override selected;
       run;
     %end;

     /****** merge overides and defaults together ******/
     proc sort data= _defaults; by year internal_bl_uid; run;
     proc sort data= _values; by year internal_bl_uid; run;
     data _valuesBl; merge _values _defaults; 
          by year internal_bl_uid; 
          if &type._override ne -1 and &type._override ne . then 
               &type._sfv_bl = &type._override;
          else &type._sfv_bl = &type._default;
          selected_&type._bl = selected;
          keep year internal_bl_uid aaf_bl &type._sfv_bl &type._agr_bl selected_&type._bl;
     run;

     /***** fill in child levels from Level-1 *****/
     %fillin_levels(data=_valuesBl,varorg=&varorg,varmeta=&varmeta,
                    type=&type,startYear=&startYear,endYear=&endYear);

     /***** merge in AGR values by businessline *****/
     data _agrs; set _agrs (rename=INTERNAL_BL_UID=uid_parent_1); run;
     proc sort data=_agrs; by uid_parent_1; run;
     proc sort data=_valuesBl; by uid_parent_1; run;
     data _valuesAgr; merge _valuesBl _agrs; by uid_parent_1;
        drop uid_parent_1; 
     run;

     proc sort data= _valuesAgr; by year internal_bl_uid; run;

     /***** merge BL values into firm *****/
     %if &sfvFirm ^= -1 %then %do;  
        data &out; merge _valuesAgr _valuesFirm; by year internal_bl_uid; run; 
     %end;
     %else %do;
        data &out; set _valuesAgr; run; 
     %end;

     /* Make sure there are no values past the global end date */
     data &out.;
        set &out.;
        where year <= &endYear.;
     run;
   %end;

   /* cleanup temporary datasets */
   proc datasets library=work memtype=data nolist nodetails;
       delete _values: _defaults: _agrs _bl1; 
   quit; 

%mend getScaleFactors;



/**************************************************************
 *  Copy values from Level-1 Business lines to all lower levels
 *  Adds a column uid_parent_1 for all rows
 **************************************************************/
%macro fillin_levels(data      =,
                     varorg    =,
                     varmeta   =,
                     startYear =,
                     endYear   =,
                     type      =);

   %local number_of_levels;

   /* How many levels are there? */
   proc sql noprint;
      select
               max(internal_bl_level)
      into
               :number_of_levels
      from
               &varorg..internal_businessline
      ;
   quit;

   /* First make sure we only have level one scale factor */
   proc sql noprint;
      create table
                    _scf_ibl_data_level_1
      as select
                    l.internal_bl_uid,
                    l.internal_bl_uid as uid_parent_1,
                    l.year,
                    l.&type._sfv_bl,
                    l.selected_&type._bl,
                    r.child_uid,
                    r.parent_uid
      from
                    &data. l
      left join
                    &varorg..internal_bl_hier r
      on
                    l.internal_bl_uid = r.child_uid
      where
                    r.parent_uid = .
      ;
   quit;

   %do i = 2 %to &number_of_levels.;
      %let j = %eval(&i. - 1);

      /* Now propogate the values down */
      proc sql noprint;
         create table
                       _scf_ibl_data_level_&i.
         as select
                       r.child_uid as internal_bl_uid,
                       l.year,
                       l.&type._sfv_bl,
                       l.selected_&type._bl,
                       l.uid_parent_1
         from
                       _scf_ibl_data_level_&j. l
         inner join
                       &varorg..internal_bl_hier r
         on
                       l.internal_bl_uid = r.parent_uid
         ;
      quit;

      /* Check that there were any observations - if not, don't need
         to continue propagating downwards */
      %orvl_ds_exists(_scf_ibl_data_level_&i., "orva_scaling_internal", false);
      %if &ds_nobs eq 0 %then %do;
         %let i = %eval(&number_of_levels. + 1);
         %let number_of_levels = &j.;
      %end;

   %end;

   data &data.;
      set
      %do i = 1 %to &number_of_levels.;
         _scf_ibl_data_level_&i.
      %end;
      ;
      drop child_uid parent_uid;
   run;

   /* cleanup temporary datasets */
   proc datasets library=work memtype=data nolist nodetails;
       delete 
         _fillinbl_: 
         _others_: 
         _scf_ibl_data:
         ; 
   quit;    
         
%mend fillin_levels;

%macro fillin_levels_firm(data     =,
                         varorg    =,
                         startYear =,
                         endYear   =,
                         type      =,
                         agr       =);

      /* make sure we have all years */
      data _years; 
         %do year = &startYear %to &endYear;
           year = &year; output;
         %end;
      run;
      proc sql;     
        create table _valuesFirm0 as select * from
        _years y left join &data v on y.year eq v.year;
      quit;

      /* add in AGR value */
      data _valuesFirm0; set _valuesFirm0;
        &type._agr_firm = &AGR;
        if aaf_firm eq . then aaf_firm = 1;
      run;

      /* add in all possible business lines and years */
      data _bl; set &varorg..internal_businessline (keep=INTERNAL_BL_UID); run;
      proc sql;     
        create table &data as select * from
        _bl left join _valuesFirm0 on 1;
      quit;
      /* cleanup temporary datasets */
      proc datasets library=work memtype=data nolist nodetails;
          delete _bl _years _valuesFirm0;
      quit;
%mend fillin_levels_firm;

/*******************************************************************************
%orvsetup(version=3.3);
options MPRINT SYMBOLGEN;
libname VARCOMP "C:\VarTablesNew\ver_3\components";
libname VARORG  "C:\VarTablesNew\ver_3\loss\org";
libname VARMETA "C:\VarTablesNew\ver_3\meta";
%orva_scaling_internal(varcomp=VARCOMP,varorg=VARORG,varmeta=VARMETA,
                       scaleUid=2166,scaleDS= scalei);
                     
********************************************************************************/
